#!/usr/bin/env python
# -*- coding: utf-8 -*-

from optparse import OptionParser
import ConfigParser
import subprocess
import os
import time
import re
import logging
import traceback
import shlex
import itertools

class Config(object):
    '''
    Предоставляет настройки из ini файла
    '''
    def __init__(self):
        self.config = ConfigParser.RawConfigParser()
        self.config.read("/linuxcash/cash/conf/graphicsmode.ini")

    def getBool(self, section, option, defaultValue):
        value = defaultValue
        try:
            value = self.config.getboolean(section, option)
        except:
            pass
        return value

    def getString(self, section, option, defaultValue):
        value = defaultValue
        try:
            value = self.config.get(section, option)
        except:
            pass
        return value

    def sections(self):
        return self.config.sections()

class Launcher(object):
    '''
    Класс для запуска приложений пользователем autologon
    '''
    def __init__(self, app):
        self.args = ['sudo', '-u', 'autologon']
        self.app = app

    def run(self, args, autologon=True):
        '''
        Запуск приложения с параметрами
        :param args - список параметров
        :return pid запущенного приложения
        '''
        args = (self.args if autologon else []) + [self.app] + args
        logging.info(u"Запуск %s: '%s'" % (self.app, " ".join(args)))
        os.environ["DISPLAY"] = ":0"
        os.environ["XAUTHORITY"] = "/home/autologon/.Xauthority"
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        logging.debug(u"PID процесса - %s" % str(p.pid))
        return p

class XRandr(Launcher):
    '''
    Взаимодействие с xrandr
    '''
    def __init__(self):
        super(XRandr, self).__init__('xrandr')

    def getAllOutputs(self):
        '''
        Все подключенные мониторы
        '''
        out = self.run(["-q"])
        return re.findall('^(.+)\sconnected', out, re.M)

    def getPrimaryOutput(self):
        '''
        Возвращает имя primary монитора или первого, если нет primary
        '''
        out = self.run(["-q"])
        outputs = re.findall('^(.+)\sprimary', out, re.M)
        if len(outputs): # есть primary-монитор
            output = (outputs[0].split(' '))
            return output[0] if len(output) else ''
        else: # нет primary-монитора - возвращаем первый попавшийся
            outputs = self.getAllOutputs()
            return outputs[0] if len(outputs) else ''
        return ''

    def query(self):
        '''
        Вывод информации о мониторах в консоль
        '''
        pgrep = subprocess.Popen(['pgrep', 'X'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = pgrep.communicate()
        if pgrep.returncode != 0:
            # если X сервер не запущен, запустим сами
            xServer = subprocess.Popen(['sudo', '-u', 'autologon', 'X'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            time.sleep(1)
            if xServer.poll() != None:
                out, err = xServer.communicate()
                print u"Ошибка старта X сервера:"
                print err
                exit(1)
            print self.run(["-q"])
            xServer.terminate()
        else:
            print self.run(["-q"])

    def getModes(self):
        """
        Функция получения списка режимов, поддерживаемых одновременно всеми интерфейсами (пересечение множеств)
        """
        ifaces = self.getInterfacesModes()

        interfaces = list()
        modes = set()
        for key in ifaces:
            if key == 'default':
                interfaces.append('default_')
            else:
                interfaces.append(key)
            if not modes:
                modes = set(ifaces[key])
            else:
                modes &= set(ifaces[key])
        modes = sorted(list(modes), key=lambda s:int(s.split('x')[0]), reverse=True)
        return modes, interfaces

    def getModesByInterface(self, interface):
        """
        Функция получения режимов для интерфейса interface
        """
        ifaces = self.getInterfacesModes()
        return ifaces[interface]

    def getInterfacesModes(self):
        """
        Функция получения всех подключеных интерфейсов и их режимов
        """
        out = self.run(['-q'])
        out = re.findall(r".*\n", out)

        ifaces = {}
        iface = ""
        for o in out:
            t = re.findall(r"^(.+)\sconnected", o)
            if t:
                iface = t[0]
                ifaces[t[0]] = []
            else:
                t = re.findall(r"   (\d+x\d+)", o)
                if t:
                    ifaces[iface].append(t[0])
        return ifaces

    def getInterfaceGeometry(self, interface):
        out = self.run(['-q'])
        outputs = re.findall('^%s.*\s(\d+)x(\d+)\+(\d+)\+(\d+)' % interface, out, re.M)
        if outputs:
            return {'width': outputs[0][0], 'height': outputs[0][1], 'x': outputs[0][2], 'y': outputs[0][3]}
        return {}

    def run(self, args):
        '''
        Запуск XRandr c параметрами
        '''
        p = super(XRandr, self).run(args)
        out, err = p.communicate()
        if p.returncode != 0:
            logging.error(str(err))
            raise Exception, "Ошибка выполнения xrandr %s:" % err
        return out

def getNextOption(config, section, options):
    '''
    Последовательное получение опций options из секции section
    Возвращает список из 2-х элементов - опция и её значение
    '''
    for option in options:
        value = config.getString(section, option, None)
        if value != None:
            yield ["--" + option, value]

def setCommondMode(xRandr, outputs, mode):
    logging.info("Установка общего режима для всех интерфейсов: " + mode)
    for output in outputs:
        try:
            xRandr.run(['--output', output, '--mode', mode])
        except Exception as e:
            logging.warning(str(e))

    return 0

def disableInterface(xRandr, inteface):
    logging.info("Отключение неиспользуемого интерфейса " + inteface)
    try:
        xRandr.run(['--output', inteface, '--off'])
    except Exception as e:
        logging.warning(str(e))

def getPrimaryInterface(config, interfaces):
    for interface in interfaces:
        if config.getString(interface, 'primary', None) == 'True':
            logging.info("Основной интерфейс: " + interface)
            return  interface

    logging.warning("Не удалось определить основной интерфейс")
    return ""     

def runApplication(xRandr, config, output):
    app = config.getString(output, 'application', None)
    if app and shlex.split(app)[:1]:
        pid = 0
        try:
            pid = Launcher(shlex.split(app)[:1][0]).run(shlex.split(app)[1:], autologon=False).pid
        except Exception as e:
            logging.error(u'Ошибка запуска приложения: %s' % repr(e))
            return 1
        geometry = xRandr.getInterfaceGeometry(output)
        Launcher('xdotool').run(
            args=['search', '--sync', '--pid', str(pid), '--name', 'unexisting_application', 'windowsize',
                  geometry['width'], geometry['height'], 'windowmove', geometry['x'], geometry['y']])

def gmMain():
    logging.basicConfig(filename='/linuxcash/logs/current/graphicsmode.log',level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(message)s')
    parser = OptionParser()
    parser.add_option("-q", "--query", action="store_true", dest="query", help=u"Вывести в консоль информацию о подключенных мониторах")
    (options, args) = parser.parse_args()

    xRandr = XRandr()
    logging.info(u"Получили XRandr")
    if options.query:
        xRandr.query()
        exit(0)

    config = Config()
    if not config.getBool('main', 'setMode', False):
        logging.warning("Установка режима выключена в настройках")
        return

    confMonitors = config.sections()
    confMonitors.remove('main')
    outputs = xRandr.getAllOutputs()
    logging.info(u"Интерфейсы в системе: %s" % outputs)

    for confMonitor in confMonitors:
        if (confMonitor not in outputs) and (confMonitor != 'default_'):
            logging.warning(u"Монитор %s заданный в настройках не подключен" % confMonitor)

    commonArgs = list(itertools.chain.from_iterable(getNextOption(config, 'main', ['fb', 'dpi', 'fbmm'])))
    commonMode = config.getString('main', 'mode', '800x600')

    optList = ['rate', 'reflect', 'rotate', 'set', 'scale', 'scale-from', 'transform', 'panning', 'gamma', 'brightness']
    
    if not confMonitors:
        return setCommondMode(xRandr, outputs, commonMode)    
    
    for output in outputs:
        if output in confMonitors or output == 'default':
            continue

        disableInterface(xRandr, output) 

    primaryInterface = getPrimaryInterface(config, outputs)

    for output in outputs:
        base = commonArgs + ['--output', output]
        args = []

        if (output in confMonitors) or (output == 'default'):
            if 'default_' in confMonitors:
                output = 'default_'
            args.append(['--mode', config.getString(output, 'mode', commonMode)])
            args.extend(list(getNextOption(config, output, optList)))
        else:
            continue
            
        primary = config.getString(output, 'primary', None) == 'True' 

        if primary:
            args.append(['--primary'])
        elif config.getString(output, 'duplicate', None) == 'True':
            logging.info("Интерфейс %s запущен в режиме дублирования %s" % (output, primaryInterface))
            args.append(['--same-as', primaryInterface])
        else:
            args.append(['--right-of', primaryInterface])

        logging.info("Настройка интерфейса " + output)
        try:
            xRandr.run(base + list(itertools.chain.from_iterable(args)))
        except Exception as e:
            logging.warning(str(e))
            for opt in args: # если пошло что-то не так и целая команда не выполнилась, то пробуем выполнить по частям
                try:
                    xRandr.run(base + opt)
                except Exception as e:
                    logging.warning(str(e))
        logging.info("Режим успешно установлен")

        if not primary:
            runApplication(xRandr, config, output)

    return 0

if __name__ == "__main__":
    try:
        gmMain()
    except:
        logging.error(traceback.format_exc())
